\KIN.XPL	21-MAR-2000	By Richard Ottosen   ottosen@idcomm.com
\
\Control program for Kinetic Art
\

\The artwork spins a pair of disks with several narrow slots cut around their
\ edges. The disks are mounted perpendicular to each other overlapping just a
\little less than the depth of the slots. A disk can only rotate if it is
\aligned into the other disks slot. Moving the disks at different speeds and
\with varying speeds creates a pleasing effect. I liken it to a high tech water
\fountain.
\
\The discs are CD's. I prefer AOL discs because they are abundant and free.
\They also have a pretty metallic look that reflects rainbows. A knibbling tool
\is used to cut the approximately 1/4 inch wide by 3/4 inch deep slots. The
\disks are made of polycarbonate (Lexan) and will withstand this treatment.
\The knibbling tool is also used to expand the center hole to mount it to a
\collet knob which is used as the hub. There can be up to 12 slots in each disk.
\
\The motors are cube shaped steppers about 1 1/2 inches on a side. They have
\unipolar windings rated for 160ma per phase at 12 volts. The particular motors
\I used were removed from old disc drives. The power supply for these steppers
\in the unregulated output of a 9 VDC, 1 amp wall transformer.
\
\The controller is a Scenix SX18AC running on its internal 4MHz clock. The five
\volts for the controller is regulated down from the 9VDC.
\
\The interface between the controller and the two motors is a ULN2803A which
\has eight darlington drivers and clamp diodes. A logic high into a driver
\switches it on to ground.
\
\There is one optical interupter sensor mounted to detect the slots in each
\disk. They are positioned 90 degrees clockwise of the overlap with the apposing
\disk. The sensor pulls to ground while a slot is seen.
\
\The control and driver circuits are constructed on a MicroEngeering Labs
\PIC Proto18 prototyping board.




\REVISIONS:
\AUG-08-96, Orignial for PIC
\JUN-03-98, Modified for Scenix
\10-MAR-2000, Made it work,  R.O.
\21-MAR-2000, Clean up comments,  R.O.

inc	CODESP;		\Intrinsics for PIC and SX microcontrollers

def	PortA=$05, PortB=$06,		\Port addresses
	IndexH=$01, IndexV=$02;		\PortA bit masks for the sensors

\Used in Sound intrinsic for delays at 4MHz instruction rate 
def	X300us=240,		\10 times faster than motor should step
	X1ms=800,		\1 millsecond per Sound cycle
	X10ms=8000;		\10ms per Sound cycle

int	PhaseH, PhaseV,		\Stepper drive phases
	SlotCntH, SlotCntV,	\Present slot number that is aligned
	Speed,			\Motor step rate
	Distance,		\Number of slots to move
	Ptr;			\Pointer for the motion table

def	Horz=0, Vert=1,		\Axis arg's
	CW=true, CCW=false,	\Rotation directions
	Slot=true, Tab=false;	\Disk features

def	Steps360= 400,		\Number of full steps in one 360-deg rotation
	Offset= 100 \(Steps360/4)\;  \How far to rotate a disk from its sensor
				     \ to align its slot with the other disk
def	MaxSlots= 12;		\Maximum number of slots allowed in a disk

\Table of positions of slots relative to initial slot:
int	MoveTbl;		\Array
char	SlotsH,	SlotsV;		\Arrays
int	NumSlotsH, NumSlotsV;	\Number of slots in a disk (measured)
int	Sections;		\Counter of sections to move.
				\ (A section is a tab and a slot).

\------------------------------------------------------------------------------

\
\Return true if the sensor detects a slot.
\
func	AtSlot(Axis);
int	Axis;
begin
\Sensor bit is low if the sensor sees a slot
if (Pin(PortA,0) & (if Axis= Horz thenIndexH else IndexV)) then
	return false else return true;
end;	\AtSlot


\
\Move the disk of one axis one direction for a number of counts.
\ The motor steps at "Speed" rate.
\
proc	Move(Axis,Dir,Cnt);
int	Axis, Dir, Cnt;
int	I;
int	HalfStep;	\Table of stepper drive pattern
begin
\FullStep:= [$66, $55, $99, $AA, $66, $55, $99, $AA];
HalfStep:= [$66, $44, $55, $11, $99, $88, $AA, $22];
\WaveStep:= [$44, $11, $88, $22, $44, $11, $88, $22];

for I:= 0, Cnt-1 do
    begin
    if Axis =Horz then
	begin
	PhaseH:= (PhaseH + (if Dir=CW then 1 else -1)) & $07;
	Pout(Pin(PortB,0)&$F0 ! HalfStep(PhaseH)&$0F, PortB, 0);
	end
    else
	begin
	PhaseV:= (PhaseV + (if Dir=CW then 1 else -1)) & $07;
	Pout(Pin(PortB,0)&$0F ! HalfStep(PhaseV)&$F0, PortB, 0);
	end;
    Sound(0, Speed, X300us);	\Delay for Speed * 300us
    end;
end;	\Move


\
\Advance to the next slot or tab while measuring the number of steps.
\ Returns the number of steps done up to the limit of "Max".
\
func	Advance(Axis,StepsMax,ToSlot);
int	Axis, Max, ToSlot; 	\True to move to slot, false to move to a tab
int	W;			\Number of stepper steps to get there

	\
	\Advance one step
	\
	proc	AdvOne;
	begin
	Move(Axis,CW,1);
	Sound(0, 50, X1ms);	\Wait 50ms to let motor quit moving 
	W:= W+1;
	end;	\AdvOne

begin	\Advance
W:= 0;
if ToSlot then
	repeat
	if AtSlot(Axis) then			\If a slot is detected
		begin
		AdvOne;				\ then step once more and make
		if AtSlot(Axis) then return W;	\ the slot is still seen
		end
		else AdvOne;			\ else only step once
	until (W >=Max)
	else
	repeat
	if not AtSlot(Axis) then	\Do the same as above for a tab
		begin
		AdvOne;
		if not AtSlot(Axis) then return W;
		end
		else AdvOne;
	until (W >=Max);
return Max;	\Return the maximum allowed
end;	\Advance


\
\Find next slot (and move to trailing edge).
\
func	FindNextSlot(Axis);
int	Axis;
int	W;	\Slot width in stepper steps
begin
\Find a tab. It may be the one already at the sensor.
W:= Advance(Axis,Steps360,Tab);
W:= Advance(Axis,Steps360,Slot);	\Move to leading edge of slot
return (Advance(Axis,Steps360,Tab));	\Move to trailing edge of slot
					\ (& get width of slot)
end;	\FindNextSlot


\
\Align the "Axis" slot with the other disk.
\
proc	Align(Axis);
int	Axis;
int	W;	\Slot width in stepper steps
begin
W:= FindNextSlot(Axis);	\Find the next slot and get its width
\
\Sensor is now at trailing edge of slot.
\Rotate back 90 deg, allowing for the extra step made after the leading edge of
\ of the slot was found
\
Move(Axis, CCW, Offset + (W/2 +1));
end;	\Align


\
\Record the number and positions of the slots in the disk.
\Exit with the first slot that was measured aligned with the other disk.
\
proc	Record(Axis);
int	Axis;
int	W, Steps;
begin
\Disk may have its slot aligned with the other disks slot, not with the sensor.
\Go to start of next tab from the unknown position.
W:= FindNextSlot(Axis);
Steps:= 0; 
if Axis = Horz then
	begin			\Record the slots in the horizontal disk
	NumSlotsH:= 0;
	while Steps <Steps360-3 do \A full rotation but not the first slot twice
		begin
		\Get width of the tab and the next slot for a total
		\ of the distance between slots
		SlotsH(NumSlotsH):= Advance(Horz,Steps360,Slot);
		W:= Advance(Horz,Steps360,Tab);
		SlotsH(NumSlotsH):= SlotsH(NumSlotsH) +W;
		Steps:= Steps +SlotsH(NumSlotsH);
		NumSlotsH:= NumslotsH+1	;	\Count the slots
		end;
	SlotCntH:= 0;	\The finish is at slot 0
	end
	else
	begin			\Record the slots in the vertical disk
	NumSlotsV:= 0;
	while Steps <Steps360-3 do \A full rotation but not the first slot twice
		begin
		\Get width of the tab and the next slot for a total
		\ of the distance between slots
		SlotsV(NumSlotsV):= Advance(Vert,Steps360,Slot);
		W:= Advance(Vert,Steps360,Tab);
		SlotsV(NumSlotsV):= SlotsV(NumSlotsV) +W;
		Steps:= Steps +SlotsV(NumSlotsV);
		NumSlotsV:= NumslotsV+1	;	\Count the slots
		end;
	SlotCntV:= 0;	\Finish at slot 0
	end;
Move(Axis, CCW, Offset +(W/2 +1));  \Align slot 0 to the other disk
end;	\Record

\==============================================================================

begin	\Main
Option($CF);		\Register $01 is RTCC =1
			\RTCC interrupt not enabled =1
			\Internal clock to RTCC =0
			\RTCC increments on low to high edge =0
			\Prescaler assigned to WDT =1
			\WDT set to maximum time =111

SlotsH:= Reserve(MaxSlots); SlotsV:= Reserve(MaxSlots);

\
\Movement table.
\
\ A postiive Speed moves the vertical disk. A positive Distance is clockwise.
\ Speed = 10 is the fastest allowed. Avoid speeds slower than the value 127.
\ Distance is in number of slots.
\ A Distance of 0 causes a pause with no movement for "Speed" times 10ms.
\ A Speed of 0 ends the table.
\
\		Dir. /Dist.	Disk /Speed	Dir /Dist.	Disk /Speed
MoveTbl:= [
		-5,		-125,		10,		100,
		-5,		-100,		-2,		150,
		3,		-50,		4,		20,
		-5,		-20,		1,		20,
		0, \Pause\	300,
		-1,		-80,		1,		50,
		-1,		-65,		1,		150,
		1,		-25,		-1,		30,
		-1,		-15,		1,		60,
		-1,		-10,		-1,		10,
		1,		-10,		-1,		10,
		-1,		-10,		1,		10,
		0, \Pause\	150,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		1,		-80,		1,		80,

\		Dir. /Dist.	Disk /Speed	Dir /Dist.	Disk /Speed
		0, \Pause\	250,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		1,		-80,		1,		80,
		0, \Pause\	400,
		-8,		-10,		8,		10,
		8,		-10,		8,		10,
		-8,		-10,		8,		10,
		8,		-10,		8,		10,
		0, \Pause\	200,
		1,		120,		1,		96,
		1,		77,		2,		61,
		2,		49,		3,		39,
		4,		31,		5,		25,
		6,		20,		7,		16,
		9,		13,		10,		12,
		11,		11,		12,		10,
		
\		Dir. /Dist.	Disk /Speed	Dir /Dist.	Disk /Speed
		0, \Pause\	300,
		-1,		-600,		1,		50,
		-1,		-400,		1,		150,
		1,		-200,		-1,		250,
		-1,		-20,		1,		500,
		-1,		-10,		-1,		10,
		1,		-10,		-1,		10,
		-1,		-10,		1,		10,
		0, \Pause\	100,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		-1,		-10,		-1,		10,
		1,		-10,		1,		10,
		1,		-80,		1,		80,
		0, \Pause\	150,
		-8,		-10,		8,		10,
		8,		-10,		8,		10,
		-8,		-10,		8,		10,
		8,		-10,		8,		10,
		0, \Pause\	600,
		0 \Dummy\,	0 \Terminator\
	];	\End of "Movetbl"

\		Dir. /Dist.	Disk /Speed	Dir /Dist.	Disk /Speed

\----------------------------------------------------------------------------
loop begin	\Main loop

\
\Re-initialize every time through the movement table in case a glitch -- such
\ as a power sag or a cat -- occurred.
\
Pout($00, PortA, 0);	\Clear Port A
Pout($13, PortA, $F);	\Set Port A direction register for bits 0 and 1 as inputs
Pout($00, PortB, 0);	\Clear Port B turning off the motors
Pout($00, PortB, $F);	\Set Port B direction register for all outputs
Sound(0, 100, X1ms);	\Wait for 100ms to let disks spring back

Speed:= 10;		\Fastest step rate
PhaseH:= 0; PhaseV:= 0;	\ Set initial phase of steppers

\
\Unjam the disks by aligning one of the horizontal disk's slots to the vertical
\ disk. Jams can be caused by ill timed shut down.
\
\Try moving a horizontal disk slot to the sensor
if  Advance(Horz,Steps360,Slot) <Steps360 then Align(Horz) \The disk is unjammed
	else
	begin				\Disk is still jammed
	\Turn off the motors to relieve the tension between the disks
	Pout($00, PortB, 0);
	Sound(0, 100, X1ms);	\Wait for 100ms to let disks spring back
	Move(Horz,CCW,3);	\Still jammed, back up a little
	end;
Align(Vert); Align(Horz);	\Get the disks and slots accurately aligned

Record(Horz);	\Find where the slots are at 
Record(Vert);	\ in each disk

\
\-----   Move the disks in a pleasing way   ------
\
Ptr:= 0;
while MoveTbl(Ptr+1) #0 do	\Do moves until the terminator is found
    begin
    Speed:= ABS(MoveTbl(Ptr+1));	\Get the speed separated from the disk
    Distance:= ABS(MoveTbl(Ptr));	\Get the distance without the direction
    if Distance =0 then Sound(0, Speed, X10ms)  \Pause for Speed * 10ms
    else
    for Sections:= 0, Distance -1 do	\Rotate by sections
	begin
     	if MoveTbl(Ptr+1) <0	\Then do horizontal motion
	    then
	    if MoveTbl(Ptr) >=0	\Then move clockwise
		then [Move(Horz, CW, SlotsH(SlotCntH));
		      SlotCntH:= SlotCntH +1;
		      if SlotCntH >=NumSlotsH then SlotCntH:= 0;
		     ]
		else [		\Move counterclockwise
		      SlotCntH:= SlotCntH -1;
	              if SlotCntH <0 then SlotCntH:= NumSlotsH -1;
	              Move(Horz, CCW, SlotsH(SlotCntH));
	              ]
	    else	\Do vertical motion
	    if MoveTbl(Ptr) >=0 \Then move clockwise
		then [Move(Vert, CW, SlotsV(SlotCntV));
		      SlotCntV:= SlotCntV +1;
		      if SlotCntV >=NumSlotsV then SlotCntV:= 0;
	             ]
		else [		\Move counterclockwise
		      SlotCntV:= SlotCntV -1;
	              if SlotCntV <0 then SlotCntV:= NumSlotsV -1;
	              Move(Vert, CCW, SlotsV(SlotCntV));
	             ];			
	end;
     Ptr:= Ptr +2;	\Next pair of entries in the movement table
     end;
end; \ Main loop -- loop forever

end;	\Main
