\SXLODI2L.XPL	MAY-09-98	Version 1.0
\I2L Loader for the Scenix Microcontroller
\Copyright (C) 1998 Loren Blaney
\
\This program converts an .I2L file into a .ASM file for the SX. This .ASM file
\ is then appended to I2L.ASM and the combination is assembled with MPASM.
\
\The .I2L file consists of printable ASCII characters interpreted as follows:
\	; = Relative load address follows (4 hex chars).
\	* = Relative address follows (4 hex chars). This address is relocated
\	    by adding the execution base address (PROGLO) to it.
\	^ = Relative address follows (4 hex chars). This address is to have the
\	    current PC value stored into it. This enables forward references.
\	After each of the above, there can follow two characters of hex for
\	each loaded byte for any number of bytes.
\	$ = End of .I2L file.
\
\Modules are not supported by this loader.
\
\The output ROM is thought of as being 8 bits wide although the physical
\ ROM is actually 12 bits wide. The first 1K of code is loaded in a
\ straightforward fashion starting at PROGLO. An extra half-K is packed
\ into the unused high nibbles of each word. After location 7FE is loaded,
\ the next byte is divided into nibbles, the high nibble is loaded at
\ PROGLO, and the low nibble is loaded at PROGLO+1.
\
\				|             |
\				| Interpreter |
\				|<--12 bits-->|
\				|    |<8 bits>|
\				+-------------+
\			PROGLO	| hi | Start  |
\				| lo | of I2L |
\				| hi | code   |
\				| lo |        |
\				|    |        |
\				|    |        |
\			   7FE	|xxxx|I2L code|
\			   7FF  | GOTO  RESET |
\				+-------------+
\
\Location 7FF is the SX's reset vector. The interpreter skips reading this
\ location, and so this loader must skip it also.
\
\LICENSE:
\This program is free software; you can redistribute it and/or modify it under
\ the terms of the GNU General Public License version 2 as published by the
\ Free Software Foundation.
\This program is distributed in the hope that it will be useful, but WITHOUT
\ ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
\ FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
\ details.
\You should have received a copy of the GNU General Public License along with
\ this program (in the file LICENSE.DOC); if not, write to the Free Software
\ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\
\You can reach me at:			Mail:	Loren Blaney
\ Email: loren_blaney@idcomm.com		502 Pine Glade Dr.
\						Nederland, CO 80466, USA
\
\This program was compiled with the version 2.3 XPL0 compiler (XPLX, 16-bit).
\ It is available from: http://www.brouhaha.com/~eric/computers/xpl0/
\ I can send you a copy of this compiler if you can't find it.
\
\REVISIONS:
\1.0, MAY-09-98, Released.


inc	C:\CXPL\CODESI;

char	ROM;		\byte array of ROM image; actual image has packed words

int	ADDR,		\address read from .I2L file (following "^")
	CHAR,		\character read from input .I2L file
	JJ,		\scratch (accumulator)
	NONHEX,		\flag: a non-hex character was read by INHEX
	PC,		\absolute execution address program counter
	PROGLO,		\absolute execution base address
	ROMINX,		\index into ROM array
	ROMSIZE;	\maximum number of bytes available for I2L code

\Variables for DOS file I/O:
def	NAMESIZE=80;	\maximum number of chars in a file name (incl path)
int	CPUREG,		\pointer to system info
	PSPSEG,		\address of PSP segment (for file I/O)
	DATASEG,	\address of data segment (heap & stack)
	INHANDLE,	\input file handle (= -1 if no input file)
	OUTHANDLE;	\output file handle (= -1 if no output file)
char	INFILENAME(NAMESIZE),	\input file name (zero terminated)
	OUTFILENAME(NAMESIZE),	\output file name
	TMPFILENAME(NAMESIZE),	\temporary output file name (.$$$)
	BAKFILENAME(NAMESIZE);	\backup output file name (.BAK)

def	BEL=$07;	\ASCII bell character

\========================== DOS I/O ROUTINES ==========================

proc	DOSOPEN(INEXT, OUTEXT);	\Open MS-DOS files
char	INEXT, OUTEXT;	\input and output extensions
int	INX;		\index into CMDLINE
char	CMDLINE($7F);	\command line tail
def	SP=$20, CR=$0D;


	proc	COPYNAME(FROM, TO, EXT);
	\Copy file name and replace any existing extension with EXT.
	\If EXT = "@@@" then use extension from FROM.
	char	FROM, TO,	\Strings are 0 terminated
		EXT;		\Extension
	int	I, J, CH;
	begin
	I:= 0;
	loop	begin
		CH:= FROM(I);
		if CH=^. ! CH=0 then quit;
		TO(I):= CH;
		if I < NAMESIZE-1 then I:= I +1;
		end;
	TO(I):= ^.;
	if I < NAMESIZE-1 then I:= I +1;
	for J:= 0, 2 do
		begin
		CH:= EXT(J) & $7F;
		TO(I):= if CH=^@ then FROM(I) else CH;
		if I < NAMESIZE-1 then I:= I +1;
		end;
	TO(I):= 0;
	end;	\COPYNAME



	proc	GETNAME(NAME);
	\Read file name from command line and return it in NAME
	\The command line is terminated by a CR.
	char	NAME;
	int	I, CH;


		proc	GETCH;	\Get character and strip out switches
		begin
		CH:= CMDLINE(INX);
		INX:= INX +1;
		if CH=^/ then
			begin			\ignore any one-char switches
			CH:= CMDLINE(INX);
			INX:= INX +1;
			if CH # CR then		\handle wise guys
				begin
				CH:= CMDLINE(INX);
				INX:= INX +1;
				end;
			end;
		end;	\GETCH


	begin
	repeat GETCH until CH # SP;
	I:= 0;
	loop	begin
		if CH=SP ! CH=CR then CH:= 0;	\terminate NAME with 0
		NAME(I):= CH;
		if CH = 0 then quit;
		if I < NAMESIZE-1 then I:= I +1;
		GETCH;
		end;
	end;	\GETNAME


begin	\DOSOPEN
BLIT(PSPSEG, $81, DATASEG, CMDLINE, $7F);

INHANDLE:= -1;   OUTHANDLE:= -1;	\set to illegal values

INX:= 0;
if INEXT(0) # SP then
	begin				\get the input file name
	GETNAME(INFILENAME);
	COPYNAME(INFILENAME, INFILENAME, INEXT);

	INHANDLE:= FOPEN(INFILENAME, 0);
	FSET(INHANDLE, ^I);
	end;

if OUTEXT(0) # SP then
	begin				\get the output file name
	if INX > 0 then INX:= INX -1;	\backup in case of CR
	GETNAME(OUTFILENAME);
	if OUTFILENAME(0) =0 then	\if no outfile then use the infile
		COPYNAME(INFILENAME, OUTFILENAME, OUTEXT)
	else	COPYNAME(OUTFILENAME, OUTFILENAME, OUTEXT);

	COPYNAME(OUTFILENAME, TMPFILENAME, "$$$");
	COPYNAME(OUTFILENAME, BAKFILENAME, "BAK");

	OUTHANDLE:= FOPEN(TMPFILENAME, 1);
	FSET(OUTHANDLE, ^O);
	end;
end;	\DOSOPEN

\-------------------------------------------------------------------------------

proc	DOSCLOSE;	\Close MS-DOS files
begin
if INHANDLE # -1 then FCLOSE(INHANDLE);

if OUTHANDLE # -1 then
	begin
	FCLOSE(OUTHANDLE);

	CPUREG(\AX\0):= $4100;		\DEL OUTFILENAME.BAK
	CPUREG(\DS\9):= DATASEG;
	CPUREG(\DX\3):= BAKFILENAME;
	SOFTINT($21);			\(.BAK file might not exist)

	CPUREG(\AX\0):= $5600;		\REN OUTFILENAME.EXT OUTFILENAME.BAK
	CPUREG(\DS\9):= DATASEG;
	CPUREG(\DX\3):= OUTFILENAME;
	CPUREG(\ES\11):= DATASEG;
	CPUREG(\DI\4):= BAKFILENAME;
	SOFTINT($21);

	CPUREG(\AX\0):= $5600;		\REN OUTFILENAME.$$$ OUTFILENAME.EXT
	CPUREG(\DS\9):= DATASEG;
	CPUREG(\DX\3):= TMPFILENAME;
	CPUREG(\ES\11):= DATASEG;
	CPUREG(\DI\4):= OUTFILENAME;
	SOFTINT($21);
	end;
end;	\DOSCLOSE

\===============================================================================

proc	WRITEROM;	\Write ROM data to .ASM output file
\Inputs: ROM,
\	 ROMINX = last location loaded into ROM (+1)
\
\				+-------------+  ---------
\			PROGLO	| hi | Start  |   |     |
\				| lo | of I2L |   |     |
\				| hi | code   |   |     |
\				| lo |        | DATAs   |
\				|    |        |   |     |
\				|    |        |   |   ROMINX
\			   7FE	|xxxx|I2L code|   |     |
\				+----+--------+  ---    |
\				     |        |         |
\		        packed	 <-- |        |         |
\				     |        |         |
\				     +--------+  ---------
\
int	I, J, DATAs, HILO, NIB;
begin
TEXT(3, "
	ORG	");   HEXOUT(3, PROGLO);   CHOUT(3, ^h);   CRLF(3);
TEXT(3, "PROGLO	EQU	$
");

DATAs:= $7FF - PROGLO;			\maximum number of DATA instructions
if ROMINX < DATAs then DATAs:= ROMINX;	\don't output unnecessary DATA instrs
for I:= 0, DATAs-1 do			\for all of the DATA instructions...
	begin
	TEXT(3, "	DATA	");	\generate DATA instruction
	J:= DATAs + I>>1;		\get high & low nibbles to pack, if any
	HILO:= if J >= ROMINX then $FF else ROM(J);
	NIB:= if I&1\odd\ then HILO&$0F\low\ else HILO>>4\high\;
	HEXOUT(3, ROM(I) ! NIB<<8);	\output regular byte and packed nibbles
	CHOUT(3, ^h);
	CRLF(3);
	end;
TEXT(3, "
	ORG	7FFh
	GOTO	RESET
	END
");
end;	\WRITEROM

\-------------------------------------------------------------------------------

proc	FATAL;		\Fatal error handler
begin
CHOUT(0, BEL);
TEXT(0, "ILLEGAL .I2L FILE
");
exit 1;
end;	\FATAL



proc	FATAL2;		\Another Fatal error handler
begin
CHOUT(0, BEL);
TEXT(0, ".I2L FILE IS TOO BIG
");
exit 1;
end;	\FATAL2



func	INHEX;
\Read an ASCII hex digit from the input .I2L file, convert it to binary, and
\ return it. Set NONHEX true if not a hex digit.
begin
CHAR:= CHIN(3);
NONHEX:= false;
case of
  CHAR>=^0 & CHAR<=^9: return CHAR - ^0;
  CHAR>=^A & CHAR<=^F: return CHAR - ^A + 10
other NONHEX:= true;
return CHAR;
end;	\INHEX



func	BYTEIN;
\Read an ASCII hex byte (two hex digits) from the input .I2L file, convert it
\ to binary, and return it. Carry is set if a non-hex digit is read.
int	I, J;
begin
I:= INHEX;
if NONHEX then return;
J:= INHEX;
if NONHEX then FATAL;
return I<<4 ! J;
end;	\BYTEIN



func	GETADDR;
\Read an ASCII hex word, convert it to binary, and return it.
int	J;
begin
J:= BYTEIN;
if NONHEX then FATAL;
J:= SWAP(J) ! BYTEIN;
if NONHEX then FATAL;
return J;
end;	\GETADDR



proc	STORE(J);	\Store J into ROM at ROMINX
int	J;
begin
if ROMINX >= ROMSIZE then FATAL2;
ROM(ROMINX):= J;
ROMINX:= ROMINX +1;
end;	\STORE



proc	STOREBYTES(J);	\Store bytes into ROM at ROMINX
int	J;
begin
loop	begin
	J:= BYTEIN;
	if NONHEX then quit;
	STORE(J);
	end;
end;	\STOREBYTES

\-------------------------------------------------------------------------------

begin	\MAIN
CPUREG:= GETREG;	\set up for file I/O
PSPSEG:= CPUREG(11);
DATASEG:= CPUREG(12);

TEXT(0, "
-- I2L LOADER FOR SX, VER 1.0 --

");
DOSOPEN("I2L", "ASM");
OPENI(3);   OPENO(3);

TEXT(0, "What is the execution base address (PROGLO)? $");
PROGLO:= HEXIN(0);

ROMSIZE:= 3*($7FF-PROGLO)/2;
ROM:= RESERVE(ROMSIZE);

TEXT(0, "Maximum I2L program length = ");   INTOUT(0, ROMSIZE);   CRLF(0);
CRLF(0);

ROMINX:= 0;
CHAR:= CHIN(3);				\get a command character
loop	begin
	case CHAR of
	  ^;:	begin
		ROMINX:= GETADDR;
		STOREBYTES;
		end;
	  ^^:	begin
		ADDR:= GETADDR;
		PC:= ROMINX + PROGLO;
		ROM(ADDR):= PC;
		ROM(ADDR+1):= SWAP(PC);
		STOREBYTES;
		end;
	  ^*:	begin
		JJ:= GETADDR + PROGLO;
		STORE(JJ);
		STORE(SWAP(JJ));
		STOREBYTES;
		end;
	  ^$:	quit
	other	begin
		CHAR:= CHIN(3);		\get a command character
		end;
	end;

\Add an exit opcode in case program tries to return (this is not currently used)
if ROMINX < ROMSIZE then STORE(0);	\if it will fit (no big deal if not)
WRITEROM;
CLOSE(3);
DOSCLOSE;
end;	\MAIN
